/* MakePri
 *
 * Convert an image in the PBM/PGM format or the BMP format (from Microsoft
 * Windows) to the "poly-raster" format.
 *
 * The poly-raster format is a simple image format that contains several
 * bitmaps; each bitmap starts with the following header:
 *    size    4     file size, including the header
 *    id      2     file signature, must be 0xa202
 *    layout  1     bit order and flags
 *    depth   1     colour depth in bits-per-pixel, 1=monochrome, 8=256 colours or grey levels, etc.
 *    width   2     image width in pixels
 *    height  2     image height in pixels
 *
 * The input file can be a PBM (monochrome bitmap) or PGM (grey-level bitmap)
 * file. Both binary and ASCII variants of PBM and PGM are supported.
 * Alternatively, Windows BMP images are supported.
 *
 * The converter currently supports only a subset of all flags.
 *      Bit     Description     Support
 *      0       column order    yes
 *      1       banded          yes
 *      2       reversed byte   yes
 *      3       planar          NO
 *      4       inverted Y-axis yes
 *      5       extended header yes
 *      6       cmap            yes (when loading a BMP image; the PGM file format does not have a colour table)
 *      7       loopback frame  yes
 *
 * The converter currently supports only the following BPP values:
 *      1       yes
 *      2       NO
 *      4       yes
 *      8       yes
 *     16       NO
 *     24       NO
 */

/*
 * Copyright (c) 2006-2010, ITB CompuPhase. All rights reserved.
 * www.compuphase.com
 *
 * Version: $Id: makepri.c 4378 2010-12-10 11:18:55Z thiadmer $
 */

#if defined __WIN32__ || defined _WIN32 || defined _Win32 || defined __NT__
  #include <windows.h>
#endif
#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "makepri.h"
#include "pri.h"


#if defined __WIN32__ || defined _WIN32 || defined _Win32 || defined __NT__
  #define DIRSEP_CHAR '\\'
#else
  #define DIRSEP_CHAR '/'
  #define stricmp(s1,s2)    strcasecmp((s1),(s2))
  typedef unsigned char   BYTE;
  typedef unsigned short  WORD;
  typedef unsigned long   DWORD;
  typedef struct tagRGBQUAD { BYTE rgbBlue, rgbGreen, rgbRed, rgbReserved; } RGBQUAD;
#endif
#if !defined _MAX_PATH
    #if defined MAX_PATH
        #define _MAX_PATH   MAX_PATH
    #elif defined MAXPATH
        #define _MAX_PATH   MAXPATH
    #else
        #define _MAX_PATH   255
    #endif
#endif

static void do_invert_y(unsigned char *image, int width, int height);


PROFILE profile_root = { NULL, 0, 0, NULL, NULL};
int profile_read(const char *filename)
{
  char label[128], description[256];
  char str[256], *ptr;
  FILE *fp;
  PROFILE *cur, *last;
  int layout, bpp, success;

  if ((fp = fopen(filename,"rt")) == 0)
    return 0;

  while (fgets(str, sizeof str, fp) != NULL) {
    /* skip leading white-space */
    for (ptr = str; *ptr != '\0' && *ptr <= ' '; ptr++)
      /* nothing */;
    /* skip empty lines and comments */
    if (*ptr == '\0' || *ptr == '#')
      continue;
    /* try to parse the line */
    description[0] = '\0';
    success = (sscanf(str, "%s %i ; %i %[^\n]", label, &layout, &bpp, description) >= 3);
    if (!success) {
      bpp = 1;
      success = (sscanf(str, "%s %i %[^\n]", label, &layout, description) >= 2);
    } /* if */
    if (success) {
      /* append to the list */
      cur = malloc(sizeof(PROFILE));
      if (cur != NULL) {
        cur->next = NULL;
        cur->layout = (unsigned short)layout;
        cur->bpp = (unsigned short)bpp;
        cur->label = strdup(label);
        cur->description = (strlen(description) > 0) ? strdup(description) : NULL;
        for (last = &profile_root; last->next != NULL; last = last->next)
          /* nothing */;
        last->next = cur;
      } /* if */
    } /* if */
  } /* while */

  fclose(fp);
  return 1;
}

void profile_free(void)
{
  PROFILE *next;

  while (profile_root.next != NULL) {
    next = profile_root.next;
    profile_root.next = next->next; /* unlink first */
    if (next->label != NULL)
      free(next->label);
    if (next->description != NULL)
      free(next->description);
    free(next);
  } /* while */
}

int profile_lookup(const char *label, unsigned char *bpp)
{
  char *ptr;
  int layout;
  unsigned char bppval;
  PROFILE *cur;

  assert(label != NULL);
  while (*label != '\0' && *label <= ' ')
    label++;
  assert(*label != '\0');

  /* first try a number or a sequence of two numbers */
  bppval = 1;
  layout = (int)strtol(label, &ptr, 0);
  if (*ptr == ';')
    bppval = (unsigned char)strtol(ptr + 1, &ptr, 0);
  if (*ptr <= ' ') {
    if (bpp != NULL)
      *bpp = bppval;
    return layout;
  } /* if */

  /* not a valid number, look up from the list of display profiles */
  for (cur = profile_root.next; cur != NULL; cur = cur->next) {
    if (cur->label != NULL && stricmp(cur->label, label) == 0)
      break;
    if (cur->description != NULL && stricmp(cur->description, label) == 0)
      break;
  } /* for */
  if (cur == NULL)
    return -1;          /* not found either */

  if (bpp != NULL)
    *bpp = cur->bpp;
  return cur->layout;
}

unsigned char invertbits1(unsigned char b)
{
  return ((b & 0x01) << 7)
       | ((b & 0x02) << 5)
       | ((b & 0x04) << 3)
       | ((b & 0x08) << 1)
       | ((b & 0x10) >> 1)
       | ((b & 0x20) >> 3)
       | ((b & 0x40) >> 5)
       | ((b & 0x80) >> 7);
}

unsigned char invertbits2(unsigned char b)
{
  return ((b & 0x03) << 6)
       | ((b & 0x0c) << 2)
       | ((b & 0x30) >> 2)
       | ((b & 0xc0) >> 6);
}

unsigned char invertbits4(unsigned char b)
{
  return ((b & 0x0f) << 4)
       | ((b & 0xf0) >> 4);
}


static unsigned char *cache_buffer = NULL;
static int cache_size, cache_top;
static unsigned char rle_count,rle_prev;

static void cache_error(const char *msg)
{
  fprintf(stderr,msg);
  abort();
}
static void cache_reset(void)
{
  if (cache_buffer==NULL) {
    cache_size=16;
    cache_buffer=malloc(cache_size*sizeof(unsigned char));
    if (cache_buffer==NULL)
      cache_error("Out of memory!\n");
  } /* if */
  cache_top=0;
  rle_count=rle_prev=0;
}
static void cache_close(FILE *fout)
{
  if (cache_buffer!=NULL) {
    if (fout!=NULL && cache_top>0)
      fwrite(cache_buffer, 1, cache_top, fout);
    free(cache_buffer);
    cache_buffer=NULL;
  } /* if */
  cache_size=0;
  cache_top=0;
}
static void cache_putc(unsigned char c)
{
  /* see whether to grow the cache */
  assert(cache_top<=cache_size);
  if ((cache_top+1)>=cache_size) {
    cache_size*=2;
    cache_buffer=realloc(cache_buffer,cache_size*sizeof(unsigned char));
    if (cache_buffer==NULL)
      cache_error("Out of memory!\n");
  } /* if */
  if (c!=rle_prev || rle_count==255) {
    cache_buffer[cache_top++]=c;
    rle_prev=c;
    rle_count=0;
  } else {
    if (rle_count==0)
      cache_buffer[cache_top++]=c;
    else
      cache_top--;
    cache_buffer[cache_top++]=rle_count++;
  } /* if */
}

static unsigned char *load_pgm(const char *filename, int *width, int *height, int *levels)
{
  FILE *fin;
  unsigned char *img;
  char line[128];
  char *ptr;
  int type, value;
  long index, size, linelimit;
  int pixels[8], pixelperbyte, pix;

  assert(filename!=NULL);
  assert(width!=NULL);
  assert(height!=NULL);
  assert(levels!=NULL);
  *width = 0;
  *height = 0;
  if ((fin=fopen(filename,"rb"))==NULL) {
    fprintf(stderr, "ERROR: failed to open file '%s' for reading\n", filename);
    return NULL;
  } /* if */

  /* read header */
  type = 0;
  *levels = 0;
  while (fgets(line,sizeof line,fin)!=NULL) {
    if ((ptr=strchr(line,'#')) != NULL)
      *ptr = '\0';
    else
      ptr = line;
    while (*ptr != '\0') {
      while (*ptr != '\0' && *ptr <= ' ')
        ptr++;
      if (*ptr != '\0') {
        if (type == 0) {
          sscanf(ptr, "P%d", &type);
          if (type == 0) {
            fprintf(stderr, "Invalid PBM/PGM/PPM image\n");
            fclose(fin);
            return NULL;
          } /* if */
          if (type == 1 || type == 4)
            *levels = 2;
        } else {
          sscanf(ptr, "%d", &value);
          if (*width == 0)
            *width = value;
          else if (*height == 0)
            *height = value;
          else if (*levels == 0)
            *levels = value + 1;
        } /* if */
      } /* if */
      while (*ptr > ' ')
        ptr++;          /* skip word */
    } /* while */
    if (*height != 0 && *levels != 0)
      break;            /* see if all header values have been read */
  } /* while */
  if ((type != 1 && type != 2 && type != 4 && type != 5)
      || *width <= 0 || *height <= 0
      || *levels <= 0 || *levels > 256)
  {
    fprintf(stderr, "Wrong or unsupported image type\n");
    fclose(fin);
    return NULL;
  } /* if */

  /* allocate memory for the pixel map */
  size = *width * *height;
  img = malloc(size * sizeof(unsigned char));
  if (img == NULL) {
    fprintf(stderr, "ERROR: insufficient memory\n");
    fclose(fin);
    return NULL;
  } /* if */

  /* read in all pixels */
  linelimit = *width;
  pixelperbyte = (type == 4) ? 8 : 1;
  for (index = 0; index < size; index += pixelperbyte) {
    if (index >= linelimit) {
      index = linelimit;
      linelimit += *width;
    } /* if */
    switch (type) {
    case 1:
      if (fscanf(fin, " %d", &pixels[0]) == 0)
        pixels[0] = EOF;
      else
        pixels[0] = !pixels[0]; /* invert black & white: 1 = white, 0 = black */
      break;
    case 2:
      if (fscanf(fin, " %d", &pixels[0]) == 0)
        pixels[0] = EOF;
      break;
    case 4:
      value = fgetc(fin);
      if (value == EOF)
        for (pix = 0; pix < 8; pix++)
          pixels[pix] = EOF;
      else
        for (pix = 0; pix < 8; pix++)
          pixels[pix] = (((0x80 >> pix) & value) != 0) ? 0 : 1;
      break;
    case 5:
      pixels[0] = fgetc(fin);
      break;
    default:
      assert(0);
    } /* switch */
    if (pixels[0] == EOF) {
      fprintf(stderr, "ERROR: truncated input file (not enough pixel data)\n");
      fclose(fin);
      free(img);
      return NULL;
    } /* if */
    for (pix = 0; pix < pixelperbyte; pix++) {
      pixels[pix] = (pixels[pix] * 255) / (*levels - 1); /* scale to 0 .. 255 */
      /* if pixelperbyte == 8 and the width is not a multiple of 8, only
       * the pixels up to the width should be stored
       */
      if (index + pix < linelimit)
        img[(int)(index + pix)] = (unsigned char)pixels[pix];
    } /* for */
  } /* for */

  fclose(fin);
  return img;
}

#if defined __WIN32__ || defined _WIN32 || defined _Win32 || defined __NT__
    #pragma -a-         /* byte packing (old Borland compilers) */
    #pragma pack(push)
    #pragma pack(1)     /* byte packing */
	#define PACKED
#else
    #define PACKED      __attribute__((packed))
#endif

typedef struct {
  char   bfSignature[2];        /* always BM */
  DWORD  bfFileSize;            /* file size */
  WORD   bfReserved[2];         /* must both be zero */
  DWORD  bfOffset;              /* file offset to bitmap data */
  /* ----- The fields beneath are part of the bitmap memory header */
  DWORD  biSize;                /* size of the memory header (40 bytes) */
  DWORD  biWidth;               /* image width */
  DWORD  biHeight;              /* image height */
  WORD   biPlanes;              /* must be 1 */
  WORD   biBitCount;            /* must be 1, 2, 4, 8 or 24 */
  DWORD  biCompression;
  DWORD  biSizeImage;           /* size of the image */
  DWORD  biXPelsPerMeter;       /* in pixels per meter? */
  DWORD  biYPelsPerMeter;       /* in pixels per meter? */
  DWORD  biClrUsed;             /* number of palette entries */
  DWORD  biClrImportant;
  /* ----- The palette */
  RGBQUAD biColors[256];
} PACKED BMPHDR;

#if defined __WIN32__ || defined _WIN32 || defined _Win32 || defined __NT__
    #pragma pack(pop)
#endif

#if !defined BI_BITFIELDS
    #define BI_BITFIELDS    3
#endif

#if !defined DIBLINESIZE
  #define DIBLINESIZE(BitCount, Width)  \
        ((((long)(Width) * (BitCount) + 31L) & ~31uL) >> 3)
#endif
#if !defined DIBHEADERSIZE
  /* size of the full header, including the colour table and the bit fields
   * the pixel data if a packed DIB starts at (LPBYTE)pbih + DIBHEADERSIZE(pbih)
   */
  #define DIBHEADERSIZE(pbih) \
          (int)( (pbih)->biSize \
               + ((pbih)->biCompression == BI_BITFIELDS ? 3 * sizeof(DWORD) : 0) \
               + ((pbih)->biBitCount <= 8 ? (pbih)->biClrUsed * sizeof(RGBQUAD) : 0))
#endif

static unsigned char *load_bmp(const char *filename, int *width, int *height, int *levels, unsigned char cmap[][3])
{
  BMPHDR header;
  FILE *fin;
  unsigned char *img;
  int sl_width, value;
  int pixels[8], pixelperbyte, pix;
  long index, size, linelimit;

  assert(filename!=NULL);
  assert(width!=NULL);
  assert(height!=NULL);
  assert(levels!=NULL);
  *width = 0;
  *height = 0;

  if ((fin=fopen(filename,"rb"))==NULL) {
    fprintf(stderr, "ERROR: failed to open file '%s' for reading\n", filename);
    return NULL;
  } /* if */

  memset(&header, 0, sizeof header);
  fread(&header, 1, sizeof header, fin);

  if (header.bfSignature[0]!='B' || header.bfSignature[1]!='M'
      || header.biPlanes!=1
      || header.biCompression!=0
      || (header.biBitCount!=1 && header.biBitCount!=4 && header.biBitCount!=8))
  {
    fprintf(stderr, "ERROR: file '%s' has an invalid or unsupported format\n", filename);
    return NULL;
  }

  *width = header.biWidth;
  *height = header.biHeight;
  *levels = 1 << header.biBitCount;
  sl_width = DIBLINESIZE(header.biBitCount, header.biWidth);
  if (header.biClrUsed == 0)
    header.biClrUsed = *levels;
  fseek(fin, DIBHEADERSIZE(&header) + 14, SEEK_SET);

  /* read in the palette */
  assert(cmap != NULL);
  if (header.biBitCount > 1 && header.biBitCount <= 8) {
    for (index = 0; index < *levels; index++) {
      cmap[index][0] = header.biColors[index].rgbRed;
      cmap[index][1] = header.biColors[index].rgbGreen;
      cmap[index][2] = header.biColors[index].rgbBlue;
    } /* for */
  } /* if */

  /* allocate memory for the pixel map */
  size = *width * *height;
  img = malloc(size * sizeof(unsigned char));
  if (img == NULL) {
    fprintf(stderr, "ERROR: insufficient memory\n");
    fclose(fin);
    return NULL;
  } /* if */

  /* read in all pixels */
  linelimit = *width;
  pixelperbyte = 8 / header.biBitCount;
  assert(pixelperbyte == 1 || pixelperbyte == 2 || pixelperbyte == 8);
  for (index = 0; index < size; index += pixelperbyte) {
    if (index >= linelimit) {
      index = linelimit;
      linelimit += *width;
      /* also skip end-of-line padding */
      for (pix = (*width + pixelperbyte - 1) / pixelperbyte; pix < sl_width; pix++)
        fgetc(fin);
    } /* if */
    switch (header.biBitCount) {
    case 1:
      value = fgetc(fin);
      if (value == EOF)
        for (pix = 0; pix < pixelperbyte; pix++)
          pixels[pix] = EOF;
      else
        for (pix = 0; pix < pixelperbyte; pix++)
          pixels[pix] = (((0x80 >> pix) & value) != 0) ? 1 : 0;
      break;
    case 4:
      value = fgetc(fin);
      if (value == EOF) {
        for (pix = 0; pix < pixelperbyte; pix++)
          pixels[pix] = EOF;
      } else {
        for (pix = 0; pix < pixelperbyte; pix++)
          pixels[pix] = (value >> 4*(pixelperbyte-pix-1)) & 0x0f;
      } /* if */
      break;
    case 8:
      pixels[0] = fgetc(fin);
      break;
    //??? 16-bpp
    //??? 24-bpp
    default:
      assert(0);
    } /* switch */
    if (pixels[0] == EOF) {
      fprintf(stderr, "ERROR: truncated input file (not enough pixel data)\n");
      fclose(fin);
      free(img);
      return NULL;
    } /* if */
    for (pix = 0; pix < pixelperbyte; pix++) {
      pixels[pix] = (pixels[pix] * 255) / (*levels - 1); /* scale to 0 .. 255 */
      /* if pixelperbyte == 8 and the width is not a multiple of 8, only
       * the pixels up to the width should be stored
       * similarly, pixels beyond the padding of the scanlines should be
       * ignored
       */
      if (index + pix < linelimit)
        img[(int)(index + pix)] = (unsigned char)pixels[pix];
    } /* for */
  } /* for */

  fclose(fin);
  return img;
}

static unsigned char *load_image(const char *filename, int *width, int *height, int *levels, unsigned char cmap[][3])
{
  unsigned char *image;
  #if defined __WIN32__ || defined _WIN32 || defined _Win32 || defined __NT__
    const char *ext;
  #endif

  #if defined __WIN32__ || defined _WIN32 || defined _Win32 || defined __NT__
    ext = strrchr(filename, '.');
    if (ext != NULL && stricmp(ext, ".bmp") == 0) {
      image = load_bmp(filename, width, height, levels, cmap);
      if (image)
        do_invert_y(image, *width, *height);
    } else {
      image = load_pgm(filename, width, height, levels);
    } /* if */
  #else
    image = load_pgm(filename, width, height, levels);
  #endif
  return image;
}

static void do_invert_y(unsigned char *image, int width, int height)
{
  int y1, y2, x;
  unsigned char t;
  unsigned char *p1, *p2;

  y1 = 0;
  y2 = height - 1;
  while (y1 < y2) {
    p1 = &image[width*y1];
    p2 = &image[width*y2];
    for (x = 0; x < width; x++) {
      t = p1[x];
      p1[x] = p2[x];
      p2[x] = t;
    } /* for */
    y1++;
    y2--;
  } /* while */
}

static void invert_y(PRHDR *hdr, unsigned char *image)
{
  /* to avoid complexity in every "save" function, we make a separate routine
   * to flip the image around its horizontal axis
   */
  if ((hdr->layout & PRLAYOUT_YINVERT) != 0)
    do_invert_y(image, hdr->width, hdr->height);
}

static void dif_images(const PRHDR *hdr1, const unsigned char *image1,
                       const PRHDR *hdr2, const unsigned char *image2,
                       BOUNDS *rect)
{
  int x, y, width, height;
  int x2, y2;
  const unsigned char *p1, *p2;

  assert(rect != NULL);
  width = (hdr1->width < hdr2->width) ? hdr1->width : hdr2->width;
  height = (hdr1->height < hdr2->height) ? hdr1->height : hdr2->height;
  rect->x = (short)width;
  rect->y = (short)height;
  x2 = 0;
  y2 = 0;

  for (y = 0; y < height; y++) {
    p1 = &image1[width*y];
    p2 = &image2[width*y];
    //??? modify this for 16-bpp and 24-bpp bitmaps
    for (x = 0; x < width; x++) {
      if (p1[x] != p2[x]) {
        if (rect->x > x)
          rect->x = (short)x;
        if (x2 < x)
          x2 = x;
        if (rect->y > y)
          rect->y = (short)y;
        if (y2 < y)
          y2 = y;
      } /* if */
    } /* for */
  } /* for */
  rect->width = (short)(x2 - rect->x + 1);
  rect->height = (short)(y2 - rect->y + 1);
}

static void adjustrect(BOUNDS *rect, const PRHDR *hdr, int hor_ver)
{
  int tmp;
  unsigned char mask;
  int pixperbyte;

  switch (hdr->depth) {
  case 1:
    mask = ~0x07;
    pixperbyte = 8;
    break;
  case 2:
    mask = ~0x03;
    pixperbyte = 4;
    break;
  case 4:
    mask = ~0x01;
    pixperbyte = 2;
    break;
  case 8:
    mask = ~0x00;
    pixperbyte = 1;
    break;
  case 16:  //???
  case 24:  //???
  default:
    assert(0);
  } /* switch */

  switch (hor_ver) {
  case 0:
    tmp = rect->x & mask;
    rect->width = (short)(((rect->x + rect->width + pixperbyte - 1) & mask) - tmp);
    rect->x = (short)tmp;
    break;
  case 1:
    tmp = rect->y & mask;
    rect->height = (short)(((rect->y + rect->height + pixperbyte - 1) & mask) - tmp);
    rect->y = (short)tmp;
    break;
  default:
    assert(0);
  } /* switch */

  /* bound width and height so that the sub-image does not exceed the main image */
  if (rect->x + rect->width > hdr->width)
    rect->width = (short)(hdr->width - rect->x);
  if (rect->y + rect->height > hdr->height)
    rect->height = (short)(hdr->height - rect->y);
}

void row_1bpp(FILE *fout, const PRHDR *hdr, const unsigned char *img, const BOUNDS *rect)
{
  int x, y;
  unsigned char b;

  /* case for 1-bpp, row-ordered */
  cache_reset();
  for (y = 0; y < rect->height; y++) {
    b = 0;
    for (x = 0; x < rect->width; x++) {
      b = (unsigned char)((b << 1) | (img[hdr->width*(y+rect->y) + x+rect->x] > 127));
      if (x % 8 == 7) {
        /* just did a full byte */
        if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
          b = invertbits1(b);
        cache_putc(b);
      } /* if */
    } /* for */
    if (rect->width % 8 != 0) {
      b <<= (8 - rect->width % 8);
      if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
        b = invertbits1(b);
      cache_putc(b);
    } /* if */
  } /* for */
  cache_close(fout);
}

void column_1bpp(FILE *fout, const PRHDR *hdr, unsigned char *img, const BOUNDS *rect)
{
  int x, y;
  unsigned char b;

  /* case for 1-bpp, column-ordered */
  cache_reset();
  for (x = 0; x < rect->width; x++) {
    b = 0;
    for (y = 0; y < rect->height; y++) {
      b = (unsigned char)((b << 1) | (img[hdr->width*(y+rect->y) + x+rect->x] > 127));
      if (y % 8 == 7) {
        /* just did a full byte */
        if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
          b = invertbits1(b);
        cache_putc(b);
      } /* if */
    } /* for */
    if (rect->height % 8 != 0) {
      b <<= (8 - rect->height % 8);
      if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
        b = invertbits1(b);
      cache_putc(b);
    } /* if */
  } /* for */
  cache_close(fout);
}

void row_band_1bpp(FILE *fout, const PRHDR *hdr, unsigned char *img, const BOUNDS *rect)
{
  int x, y, y2;
  unsigned char b;

  /* case for 1-bpp, banded */
  cache_reset();
  for (y = 0; y < rect->height; y += 8) {
    b = 0;
    for (x=0; x < rect->width; x++) {
      for (y2 = 0; y2 < 8; y2++) {
        b = (unsigned char)(b << 1);
        if (y + rect->y + y2 < hdr->height)
          b = (unsigned char)(b | (img[hdr->width*(y+rect->y+y2) + x+rect->x] > 127));
      } /* for */
      if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
        b = invertbits1(b);
      cache_putc(b);
    } /* for */
  } /* for */
  cache_close(fout);
}

void column_band_1bpp(FILE *fout, const PRHDR *hdr, unsigned char *img, const BOUNDS *rect)
{
  int x, y, x2;
  unsigned char b;

  /* case for 1-bpp, column-ordered bands */
  cache_reset();
  for (x = 0; x < rect->width; x += 8) {
    b = 0;
    for (y = 0; y < rect->height; y++) {
      for (x2 = 0; x2 < 8; x2++) {
        b = (unsigned char)(b << 1);
        if (x + rect->x + x2 < hdr->width)
          b = (unsigned char)(b | (img[hdr->width*(y+rect->y) + (x+rect->x)+x2] > 127));
      } /* for */
      if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
        b = invertbits1(b);
      cache_putc(b);
    } /* for */
  } /* for */
  cache_close(fout);
}

void row_4bpp(FILE *fout, const PRHDR *hdr, const unsigned char *img, const BOUNDS *rect)
{
  int x, y;
  unsigned char b;

  /* case for 4-bpp, row-ordered */
  cache_reset();
  for (y = 0; y < rect->height; y++) {
    b = 0;
    for (x = 0; x < rect->width; x++) {
      b = (unsigned char)((b << 4) | ((img[hdr->width*(y+rect->y) + x+rect->x] >> 4) & 0x0f));
      if (x % 2 == 1) {
        /* just did a full byte */
        if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
          b = invertbits4(b);
        cache_putc(b);
      } /* if */
    } /* for */
    if (rect->width % 2 != 0) {
      b <<= 4;
      if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
        b = invertbits4(b);
      cache_putc(b);
    } /* if */
  } /* for */
  cache_close(fout);
}

void column_4bpp(FILE *fout, const PRHDR *hdr, unsigned char *img, const BOUNDS *rect)
{
  int x, y;
  unsigned char b;

  /* case for 4-bpp, column-ordered */
  cache_reset();
  for (x = 0; x < rect->width; x++) {
    b = 0;
    for (y = 0; y < rect->height; y++) {
      b = (unsigned char)((b << 4) | ((img[hdr->width*(y+rect->y) + x+rect->x] >> 4) & 0x0f));
      if (y % 2 == 1) {
        /* just did a full byte */
        if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
          b = invertbits4(b);
        cache_putc(b);
      } /* if */
    } /* for */
    if (rect->height % 2 != 0) {
      b <<= 4;
      if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
        b = invertbits4(b);
      cache_putc(b);
    } /* if */
  } /* for */
  cache_close(fout);
}

void row_band_4bpp(FILE *fout, const PRHDR *hdr, unsigned char *img, const BOUNDS *rect)
{
  int x, y, y2;
  unsigned char b;

  /* case for 4-bpp, banded */
  cache_reset();
  for (y = 0; y < rect->height; y += 2) {
    b = 0;
    for (x=0; x < rect->width; x++) {
      for (y2 = 0; y2 < 2; y2++) {
        b = (unsigned char)(b << 4);
        if (y + rect->y + y2 < hdr->height)
          b = (unsigned char)(b | ((img[hdr->width*(y+rect->y+y2) + x+rect->x] >> 4) & 0x0f));
      } /* for */
      if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
        b = invertbits4(b);
      cache_putc(b);
    } /* for */
  } /* for */
  cache_close(fout);
}

void column_band_4bpp(FILE *fout, const PRHDR *hdr, unsigned char *img, const BOUNDS *rect)
{
  int x, y, x2;
  unsigned char b;

  /* case for 4-bpp, column-ordered bands */
  cache_reset();
  for (x = 0; x < rect->width; x += 2) {
    b = 0;
    for (y = 0; y < rect->height; y++) {
      for (x2 = 0; x2 < 2; x2++) {
        b = (unsigned char)(b << 4);
        if (x + rect->x + x2 < hdr->width)
          b = (unsigned char)(b | ((img[hdr->width*(y+rect->y) + x+rect->x+x2] >> 4) & 0x0f));
      } /* for */
      if ((hdr->layout & PRLAYOUT_REVERSE) != 0)
        b = invertbits4(b);
      cache_putc(b);
    } /* for */
  } /* for */
  cache_close(fout);
}

void row_8bpp(FILE *fout, const PRHDR *hdr, const unsigned char *img, const BOUNDS *rect)
{
  int x, y;

  /* case for 8-bpp, row-ordered, not planar */
  cache_reset();
  for (y = 0; y < rect->height; y++) {
    for (x = 0; x < rect->width; x++)
      cache_putc(img[hdr->width*(y+rect->y) + x+rect->x]);
  } /* for */
  cache_close(fout);
}

void column_8bpp(FILE *fout, const PRHDR *hdr, unsigned char *img, const BOUNDS *rect)
{
  int x, y;

  /* case for 8-bpp, column-ordered, not planar */
  cache_reset();
  for (x = 0; x < rect->width; x++) {
    for (y = 0; y < rect->height; y++)
      cache_putc(img[hdr->width*(y+rect->y) + x+rect->x]);
  } /* for */
  cache_close(fout);
}

int main(int argc, char *argv[])
{
  int x, y, n;
  int type, levels, argbase;
  unsigned char *img, *imgprev;
  PRHDR hdr, hdrprev;
  PREXTHDR exthdr;
  BOUNDS rect;
  FILE *fout;
  unsigned char bpp;
  long fbase;
  char path[_MAX_PATH];
  char *ptr;
  unsigned char cmap[256][3];

  if (argc < 4 || argc % 2 != 0) {
    #if defined __WIN32__ || defined _WIN32 || defined _Win32 || defined __NT__
      printf("makepri " VERSION_STR ": convert Windows Bitmap, PBM and PGM images to poly-raster format\n\n");
    #else
      printf("makepri " VERSION_STR ": convert PBM and PGM images to poly-raster format\n\n");
    #endif
    printf("Usage: makepri <output> <type> <input> [<type> <input> [...]]\n\n");
    printf("The image types are decimal or hexadecimal values that specify the lay-out of\n"
           "the bitmap to store (in the polyraster file). Alternatively, you may use a\n"
           "device name, like \"gu372\". This device is then looked up from the file\n"
           "\"displays.prf\".\n\n");
    printf("You may add multiple input files with different lay-outs into a single output\n"
           "image file. In fact, that's the point of it being called \"poly\"-raster.\n\n");
    printf("To create animated poly-raster images, use a \"+\" instead of a type value; for\n"
           "example, the command:\n"
           "\tmakepri anim.pri ks0108 frame1.pgm + frame2.pgm + frame3.pgm\n"
           "creates an animation of three frames for the KS0108 LCD controller. The spaces\n"
           "around the \"+\" symbols are required.\n");
    return 1;
  } /* if */

  /* locate and open the database file */
  strcpy(path, argv[0]);
  ptr = strrchr(path, DIRSEP_CHAR);
  #if defined __WIN32__ || defined _WIN32 || defined _Win32 || defined __NT__
    if (ptr == NULL)
      ptr = strchr(path, ':');
  #endif
  if (ptr == NULL)
    path[0] = '\0';
  else
    *(ptr + 1) = '\0';
  strcat(path, "displays.prf");
  if (!profile_read(path))
    fprintf(stderr, "File displays.prf is not found\n");

  /* write the output file */
  fout=fopen(argv[1],"w+b");
  if (fout==NULL) {
    fprintf(stderr, "ERROR: failed to open file '%s' for writing\n",argv[2]);
    return 1;
  } /* if */

  /* open source image */
  argbase = 2;
  img = NULL;
  imgprev = NULL;
  memset(&hdr, 0, sizeof hdr);  /* to avoid compiler warnings */
  do {
    if (imgprev != NULL)
      free(imgprev);
    if (argv[argbase][0] == '+') {
      imgprev = img;
      hdrprev = hdr;
    } else {
      type = profile_lookup(argv[argbase], &bpp);
      imgprev = NULL;
      if (img != NULL)
        free(img);
    } /* if */
    if (type < 0) {
      fprintf(stderr, "Display type \"%s\" not found (or invalid image type number)\n", argv[argbase]);
      return 1;
    } /* if */
    memset(cmap, 0, sizeof cmap);
    img = load_image(argv[argbase + 1], &x, &y, &levels, cmap);
    if (img == NULL)
      return 1;   /* error message already given */

    if (x <= 0 || y <= 0) {
      fprintf(stderr, "ERROR: invalid size (width=%d, height=%d)\n", x, y);
      free(img);
      fclose(fout);
      return 1;
    } /* if */

    /* compare images (if this is an animation) */
    if (imgprev != NULL) {
      dif_images(&hdr, img, &hdrprev, imgprev, &rect);
      /* case for 1-bpp, column-ordered */
      switch (hdr.layout & 0x03) {
      case 0x00:
      case 0x03:
        adjustrect(&rect, &hdr, 0);
        break;
      case 0x01:
      case 0x02:
        adjustrect(&rect, &hdr, 1);
        break;
      } /* switch */
    } else {
      rect.x = 0;
      rect.y = 0;
      rect.width = (unsigned short)x;
      rect.height = (unsigned short)y;
    } /* if */

    /* build & write header (but not all fields are final) */
    fbase = ftell(fout);  /* save base */
    hdr.id = PRSIGNATURE;
    hdr.width = (unsigned short)x;
    hdr.height = (unsigned short)y;
    hdr.depth = bpp;
    hdr.layout = (unsigned short)type;
    hdr.size = ((hdr.width + 7) / 8 * hdr.height * hdr.depth) + sizeof(PRHDR);

    /* erase irrelevant flags */
    if (hdr.depth == 1 || hdr.depth == 8)
      hdr.layout &= ~PRLAYOUT_PLANAR;
    if (hdr.depth == 8)
      hdr.layout &= ~PRLAYOUT_REVERSE;
    if (hdr.depth >= 8)
      hdr.layout &= ~PRLAYOUT_BANDED;
    /* set other flags */
    if (imgprev != NULL) {
      hdr.layout |= PRLAYOUT_EXTHDR;
      hdr.size += sizeof(PREXTHDR);
    } /* if */
    if (argbase > 2 && argbase + 2 >= argc)
      hdr.layout |= PRLAYOUT_LOOPBACK;

    fwrite(&hdr, sizeof hdr, 1, fout);

    /* write the extended header */
    if (imgprev != NULL) {
      exthdr.dx = rect.x;
      exthdr.dy = rect.y;
      exthdr.delay = 0;
      fwrite(&exthdr, sizeof exthdr, 1, fout);
    } /* if */

    /* write colour table (unless it wasn't loaded) */
    for (n = 0; n < (1 << bpp); n++)
      if (cmap[n][0] != 0 || cmap[n][1] != 0 || cmap[n][2] != 0)
        break;
    if (n < (1 << bpp)) {
      hdr.layout |= PRLAYOUT_CMAP;
      fwrite(cmap, 3, 1 << bpp, fout);
    } /* if */

    /* perform some standard operations on the image */
    invert_y(&hdr, img);

    switch (hdr.depth) {
    case 1:
      /* case for 1-bpp */
      switch (hdr.layout & 0x03) {
      case 0x00:
        row_1bpp(fout, &hdr, img, &rect);
        break;
      case 0x01:
        column_1bpp(fout, &hdr, img, &rect);
        break;
      case 0x02:
        row_band_1bpp(fout, &hdr, img, &rect);
        break;
      case 0x03:
        column_band_1bpp(fout, &hdr, img, &rect);
        break;
      } /* switch */
      break;
    case 8:
      /* case for 8-bpp */
      switch (hdr.layout & 0x0f) {
      case 0x00:
        row_8bpp(fout, &hdr, img, &rect);
        break;
      case 0x01:
        column_8bpp(fout, &hdr, img, &rect);
        break;
      default:
        fprintf(stderr, "ERROR: unsupported layout for 8-bpp colour depth\n");
      } /* switch */
      break;
    default:
      fprintf(stderr, "ERROR: unsupported colour depth\n");
    } /* switch */

    /* update the header and calculate the checksum */
    hdr.size = ftell(fout) - fbase;
    hdr.width = rect.width;
    hdr.height = rect.height;
    fseek(fout, fbase, SEEK_SET);
    fwrite(&hdr, sizeof hdr, 1, fout);
    fseek(fout, 0, SEEK_END);

    hdr.width = (unsigned short)x;
    hdr.height = (unsigned short)y;

    argbase += 2;
  } while (argbase < argc);

  if (img != NULL)
    free(img);
  if (imgprev != NULL)
    free(imgprev);
  profile_free();
  fclose(fout);
  return 0;
}
